'use strict';
/*
 * Copyright (c) 2017 Houston Inc. Consulting Oy
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

const Connectors = require('loopback-connector');
const DocumentClient = require('documentdb').DocumentClient; // NB: DocumentDB just changed its name to CosmosDB.
const inherits = require('util').inherits;
const debug = require('debug')('loopback:connector:cosmosdb');

exports.initialize = function(dataSource, callback) {
    if (!DocumentClient) {
        return;
    }

    const settings = dataSource.settings;

    dataSource.connector = new CosmosDB(settings);

    if (callback) {
        dataSource.connector.connect(callback);
    }
};

function CosmosDB(settings) {
    Connectors.Connector.call(this, 'cosmosdb', settings);

    this.host = settings.host;
    this.masterKey = settings.masterKey;
    this.databaseName = settings.databaseName || undefined;
    this.collectionName = settings.collectionName || undefined;
    this.client = null;
    this.queryOptions = {
        enableCrossPartitionQuery: settings.enableCrossPartitionQueries
    };

    if (this.databaseName !== undefined && this.collectionName !== undefined) {
        debug('Selected %j collection in %j database', this.collectionName, this.databaseName);
    }

    // Initialize a retry policy.
    if (settings.retry !== null && settings.retry !== undefined) {
        this.retryPolicy = {
            max: settings.retry.max
        };
    }
}

inherits(CosmosDB, Connectors.Connector);

CosmosDB.prototype.connect = function(callback) {
    if (this.databaseName === undefined) {
        return callback(new Error('No database selected, it must be provided in dataSource.settings'));
    }

    if (this.collectionName === undefined) {
        return callback(new Error('No collection selected, it must be provided in dataSource.settings'));
    }

    debug('Connecting to %j with retry policy %j', this.host, this.retryPolicy);

    this.client = new DocumentClient(this.host, { masterKey: this.masterKey });
    callback(null);
};

CosmosDB.prototype.disconnect = function(callback) {
    debug('Disconnecting from %j', this.host);
    this.client = null;
    callback(null);
};

CosmosDB.prototype.create = function(model, data, options, callback) {
    const self = this;

    this.client.createDocument(buildCollectionUri(this.databaseName, this.collectionName), data, function(error, document) {
        if (error) {
            return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                self.create(model, data, Object.assign({
                    retry: retries
                }), callback)
            );
        }

        debug('%j document created in %j (%j)', document.id, self.collectionName, self.databaseName);
        callback(null, document.id);
    });
};

CosmosDB.prototype.updateOrCreate = function(model, data, options, callback) {
    const modelDefinition = this.getModelDefinition(model);
    const self = this;

    this.client.upsertDocument(buildCollectionUri(this.databaseName, this.collectionName), data, function(error, document) {
        if (error) {
            return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                self.updateOrCreate(model, data, Object.assign(options, {
                    retry: retries
                }), callback)
            );
        }

        debug('%j document updated in %j (%j)', document.id, self.collectionName, self.databaseName);
        callback(null, dropNonViewProperties(modelDefinition, document));
    });
};

CosmosDB.prototype.replaceOrCreate = CosmosDB.prototype.updateOrCreate;

CosmosDB.prototype.replaceById = function(model, id, data, options, callback) {
    if (id === null || id === undefined) {
        return callback(new Error('ID value is required'));
    }

    const modelDefinition = this.getModelDefinition(model);
    const self = this;

    this.client.replaceDocument(buildDocumentUri(this.databaseName, this.collectionName, id), data, function(error, document) {
        if (error) {
            return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                self.replaceById(model, id, data, Object.assign(options, {
                    retry: retries
                }), callback)
            );
        }

        debug('%j document replaced in %j (%j)', document.id, self.collectionName, self.databaseName);
        callback(null, dropNonViewProperties(modelDefinition, data));
    });
};

CosmosDB.prototype.all = function(model, filter, options, callback) {
    const collectionUri = buildCollectionUri(this.databaseName, this.collectionName);
    const modelDefinition = this.getModelDefinition(model);
    const self = this;

    try {
        const querySpec = buildQuerySpecForModel(modelDefinition, filter);
        const iterator = this.client.queryDocuments(collectionUri, querySpec, this.queryOptions);

        iterator.toArray(function(error, documents) {
            if (error) {
                return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                    self.all(model, filter, Object.assign(options, {
                        retry: retries
                    }), callback)
                );
            }

            callback(null, documents.map(x => dropNonViewProperties(modelDefinition, x)));
        });
    } catch (error) {
        callback(error);
    }
};

CosmosDB.prototype.save = function(model, data, options, callback) {
    const self = this;

    this.updateOrCreate(model, data, options, function(error, id) {
        if (error) {
            return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                self.save(model, data, Object.assign(options, {
                    retry: retries
                }), callback)
            );
        }

        data.id = id;
        callback(null, data);
    });
};

CosmosDB.prototype.count = function(model, where, options, callback) {
    const collectionUri = buildCollectionUri(this.databaseName, this.collectionName);
    const self = this;

    try {
        const querySpec = buildQuerySpecForModel(this.getModelDefinition(model), { where: where }, ['COUNT(c._self) AS count']);
        const iterator = this.client.queryDocuments(collectionUri, querySpec);

        iterator.nextItem(function(error, result) {
            if (error) {
                return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                    self.count(model, where, Object.assign(options, {
                        retry: retries
                    }), callback)
                );
            }

            callback(null, result.count);
        });
    } catch (error) {
        callback(error);
    }
};

CosmosDB.prototype.update = function(model, where, data, options, callback) {
    const collectionUri = buildCollectionUri(this.databaseName, this.collectionName);
    const modelDefinition = this.getModelDefinition(model);
    const self = this;

    try {
        // Translate data properties from view to DB.
        data = translateDBObjectFromView(modelDefinition, data);

        const querySpec = buildQuerySpecForModel(modelDefinition, { where: where });
        const iterator = this.client.queryDocuments(collectionUri, querySpec);

        let totalItems = 0;
        const failures = [];

        // Recursively apply update operations to all matching documents.
        // Underlying client library makes sure that all documents are
        // fetched. We just travel through the query iterator and execute
        // replace for each document in sequence. Because CosmosDB has no
        // multi-document transactions, we do not stop on individual
        // errors but rather execute the whole query iterator until end.
        // At the end, we report the ratio of how many operations succeeded
        // to caller. So that they can decide what to do.
        const next = () => {
            iterator.nextItem((error, document) => {
                if (error) {
                    retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                        self.update(model, where, data, Object.assign(options, {
                            retry: retries
                        }), callback)
                    );
                }

                // Execute replace to next item in iteration.
                else if (document !== undefined) {
                    ++totalItems;

                    // Set updated properties to document.
                    Object.assign(document, data);

                    this.replaceById(model, document.id, document, options, function(error) {
                        if (error) {
                            debug('Individual replace operator failed: %j', error);
                            failures.push(document.id);
                        }

                        // Just move to next item, ignore possible errors.
                        next();
                    });
                }

                // Finish up the operation and report the caller.
                else {
                    const numberOfSuccesses = totalItems - failures.length;

                    debug('Batch update success rate is %j', numberOfSuccesses / totalItems);

                    callback(null, {
                        count: numberOfSuccesses,
                        successRate: numberOfSuccesses / totalItems,
                        failures: failures
                    });
                }
            });
        };

        next();
    } catch (error) {
        callback(error);
    }
};

CosmosDB.prototype.destroyAll = function(model, where, options, callback) {
    const collectionUri = buildCollectionUri(this.databaseName, this.collectionName);
    const self = this;

    try {
        const querySpec = buildQuerySpecForModel(this.getModelDefinition(model), { where: where }, ['c.id']);
        const iterator = this.client.queryDocuments(collectionUri, querySpec);

        let totalItems = 0;
        const failures = [];

        // Recursively apply delete operations to all matching documents.
        // Underlying client library makes sure that all documents are
        // fetched. We just travel through the query iterator and execute
        // delete for each document in sequence. Because CosmosDB has no
        // multi-document transactions, we do not stop on individual
        // errors but rather execute the whole query iterator until end.
        // At the end, we report the ratio of how many operations succeeded
        // to caller. So that they can decide what to do.
        const next = () => {
            iterator.nextItem((error, document) => {
                if (error) {
                    retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                        self.destroyAll(model, where, Object.assign(options, {
                            retry: retries
                        }), callback)
                    );
                }

                // Execute delete to next item in iteration.
                else if (document !== undefined) {
                    ++totalItems;

                    self.destroyOne(model, document.id, options, function (error) {
                        if (error) {
                            debug('Individual delete operator failed: %j', error);
                            failures.push(document.id);
                        }

                        // Just move to next item, ignore possible errors.
                        next();
                    });
                }

                // Finish up the operation and report the caller.
                else {
                    const numberOfSuccesses = totalItems - failures.length;

                    debug('Batch destroy success rate is %j', numberOfSuccesses / totalItems);

                    callback(null, {
                        count: numberOfSuccesses,
                        successRate: numberOfSuccesses / totalItems,
                        failures: failures
                    });
                }
            });
        };

        next();
    } catch (error) {
        callback(error);
    }
};

CosmosDB.prototype.destroyOne = function(model, id, options, callback) {
    const documentUri = buildDocumentUri(this.databaseName, this.collectionName, id);
    const self = this;

    this.client.deleteDocument(documentUri, function (error) {
        if (error) {
            return retryOrFail(self.retryPolicy, error, options.retry, callback, retries =>
                self.destroyOne(model, id, Object.assign(options, {
                    retry: retries
                }), callback)
            );
        }

        debug('%j document deleted', id);
        callback(null);
    });
};

CosmosDB.prototype.updateAttributes = function(model, id, data, options, cb) {
    // CosmosDB has no native support for individual attribute updates.
    // Instead we just update documents with given ID and replace them
    // with newer versions.
    this.update(model, { id: id }, data, options, cb);
};

function buildCollectionUri(databaseName, collectionName) {
    return `dbs/${databaseName}/colls/${collectionName}`;
}

function buildDocumentUri(databaseName, collectionName, documentId) {
    return `dbs/${databaseName}/colls/${collectionName}/docs/${documentId}`;
}

function dropNonViewProperties(modelDefinition, object) {
    const viewObject = {};

    for (const dbProperty in object) {
        try {
            const viewProperty = translateViewPropertyFromDB(modelDefinition, dbProperty);

            viewObject[viewProperty] = object[dbProperty];
        } catch (error) {
            // Property is dropped because it cannot be translated to view property.
        }
    }

    return viewObject;
}

function translateDBPropertyFromView(modelDefinition, viewProperty) {
    if (modelDefinition.properties[viewProperty] !== undefined && modelDefinition.properties[viewProperty].cosmosdb !== undefined && modelDefinition.properties[viewProperty].cosmosdb.propertyName !== undefined) {
        return modelDefinition.properties[viewProperty].cosmosdb.propertyName;
    }

    if (modelDefinition.properties[viewProperty] !== undefined) {
        return viewProperty;
    }

    throw new Error(`'${viewProperty}' is not any of available model properties: ${Object.keys(modelDefinition.properties).join(', ')}, or it doesn't have a valid 'cosmosdb.propertyName' configuration.`);
}

function translateViewPropertyFromDB(modelDefinition, dbProperty) {
    for (const property in modelDefinition.properties) {
        if (modelDefinition.properties[property].cosmosdb !== undefined && modelDefinition.properties[property].cosmosdb.propertyName === dbProperty) {
            return property;
        }
    }

    if (modelDefinition.properties[dbProperty] !== undefined) {
        return dbProperty;
    }

    throw new Error(`'${dbProperty}' is not any of available model properties: ${Object.keys(modelDefinition.properties).join(', ')}, or it doesn't have a valid 'cosmosdb.propertyName' configuration.`);
}

function translateDBObjectFromView(modelDefinition, object) {
    const dbObject = {};

    for (const viewProperty in object)  {
        const dbProperty = translateDBPropertyFromView(modelDefinition, viewProperty);

        dbObject[dbProperty] = object[viewProperty];
    }

    return dbObject;
}

function isCaseInsensitive(modelDefinition, property) {
    return modelDefinition.properties[property] !== undefined && modelDefinition.properties[property].cosmosdb !== undefined && !!modelDefinition.properties[property].cosmosdb.caseInsensitive;
}

function buildWhereClauses(modelDefinition, params, where) {
    return Object.keys(where).map(x => {
        const normalizedKey = x.toUpperCase().trim();

        // Build a top-level logical operator.
        if (['AND', 'OR'].indexOf(normalizedKey) >= 0) {
            // All sub-level logical operators are AND if nothing else is specified.
            const logicalClause = where[x]
                .map(y => buildWhereClauses(modelDefinition, params, y))
                .filter(y => !!y)
                .join(` ${normalizedKey} `);

            if (logicalClause.length > 0) {
                return `(${logicalClause})`;
            }

            // No logical clause.
            return undefined;
        }

        const dbProperty = translateDBPropertyFromView(modelDefinition, x);

        // Use CONTAINS() function to build LIKE operator since CosmosDB does not
        // support LIKEs.
        if (where[x]['like'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['like']));

            return `CONTAINS(${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))}, @_${params.length.toString()})`;
        }

        // Use CONTAINS() function to build NOT LIKE operator since CosmosDB does not
        // support LIKEs.
        if (where[x]['nlike'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['nlike']));

            return `NOT CONTAINS(${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))}, @_${params.length.toString()})`;
        }

        // Use CONTAINS() and LOWER() functions to build case-insensitive LIKE
        // operator since CosmosDB does not support case-insensitive LIKEs.
        if (where[x]['ilike'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, (where[x]['ilike'] || '').toLowerCase()));

            return `CONTAINS(LOWER(${escapeColumn('c', dbProperty)}), @_${params.length})`;
        }

        // Use CONTAINS() and LOWER() functions to build case-insensitive NOT LIKE
        // operator since CosmosDB does not support case-insensitive LIKEs.
        if (where[x]['nilike'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, (where[x]['nilike'] || '').toLowerCase()));

            return `NOT CONTAINS(LOWER(${escapeColumn('c', dbProperty)}), @_${params.length})`;
        }

        // Build greater than operator.
        if (where[x]['gt'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['gt']));

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} > @_${params.length.toString()}`;
        }

        // Build greater than or equal operator.
        if (where[x]['gte'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['gte']));

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} >= @_${params.length.toString()}`;
        }

        // Build less than operator.
        if (where[x]['lt'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['lt']));

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} < @_${params.length.toString()}`;
        }

        // Build less than or equal operator.
        if (where[x]['lte'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['lte']));

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} <= @_${params.length.toString()}`;
        }

        // Build IN operator.
        if (where[x]['inq'] !== undefined) {
            const positions = [];

            for (let i = 0; i < where[x]['inq'].length; ++i) {
                params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['inq'][i]));
                positions.push(params.length);
            }

            const inParams = positions.map(i => `@_${i}`);

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} IN (${inParams.join(',')})`;
        }

        // Build NOT IN operator.
        if (where[x]['nin'] !== undefined) {
            const positions = [];

            for (let i = 0; i < where[x]['nin'].length; ++i) {
                params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['nin'][i]));
                positions.push(params.length);
            }

            const inParams = positions.map(i => `@_${i}`);

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} NOT IN (${inParams.join(',')})`;
        }

        // Build non-equality operator.
        if (where[x]['neq'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['neq']));

            return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} <> @_${params.length.toString()}`;
        }

        // Build BETWEEN operator.
        if (where[x]['between'] !== undefined) {
            if (where[x]['between'].length !== 2) {
                throw new Error(`'between' operator has incorrect number of parameters. It should have exactly 2, but has ${where[x]['between'].length}.`);
            }

            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['between'][0]));
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['between'][1]));

            return `(${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} BETWEEN @_${params.length - 1} AND @_${params.length.toString()})`;
        }

        // Use ARRAY_CONTAINS to build any operator.
        if (where[x]['any'] !== undefined) {
            params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]['any']));

            return `ARRAY_CONTAINS(${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))}, @_${params.length.toString()})`;
        }

        // By default, assume equality operator.
        params.push(decorateQueryParameter(modelDefinition, dbProperty, where[x]));

        return `${decoreateDBQueryValue(modelDefinition, dbProperty, escapeColumn('c', dbProperty))} = @_${params.length.toString()}`;
    }).filter(x => x !== undefined)

      // All clauses are AND if nothing else is specified.
      .join(' AND ');
}

function decoreateDBQueryValue(modelDefinition, dbProperty, value) {
    if (isCaseInsensitive(modelDefinition, dbProperty)) {
        return `LOWER(${value})`;
    }

    return value;
}

function decorateQueryParameter(modelDefinition, dbProperty, value) {
    if (isCaseInsensitive(modelDefinition, dbProperty)) {
        return value.toLowerCase();
    }

    return value;
}

function escapeColumn(row, dbProperty) {
    return `${row}["${dbProperty}"]`;
}

function buildQuerySpecForModel(modelDefinition, filter, select, orderBy) {
    filter = filter || {};

    const modelProperties = Object.keys(modelDefinition.properties);
    const querySelect = select || modelProperties.map(x => escapeColumn('c', translateDBPropertyFromView(modelDefinition, x)));

    // Set default ordering.
    let queryOrderBy = orderBy || ['c._ts'];

    // Build ordering if it is set in filters.
    if (filter.order) {
        filter.order = Array.isArray(filter.order) ? filter.order : [filter.order];

        queryOrderBy = (filter.order || []).map(x => {
            const order = x.split(' ', 2);
            const dbProperty = translateDBPropertyFromView(modelDefinition, order[0]);

            // Normalize order by type if given.
            if (order.length > 1) {
                order[1] = order[1].toUpperCase().trim();
            }

            // Set default order by type.
            else {
                order.push('ASC');
            }

            if (['ASC', 'DESC'].indexOf(order[1]) < 0) {
                throw new Error(`Order by '${order[1]}' is not allowed for the field '${order[0]}'.`);
            }

            return `${escapeColumn('c', dbProperty)} ${order[1]}`;
        });
    }

    const queryParams = [];
    const queryWhere = buildWhereClauses(modelDefinition, queryParams, filter.where || {});

    const querySpec = {
        query: `SELECT${isFinite(filter.limit) ? ' TOP ' + filter.limit : ''} ${querySelect.join(',')} FROM c ${queryWhere.length > 0 ? 'WHERE ' + queryWhere : ''} ORDER BY ${queryOrderBy.join(',')}`,
        parameters: queryParams.map((x, i) => ({
            name: `@_${i + 1}`,
            value: x
        }))
    };

    debug('SQL: %j, params: %j', querySpec.query, querySpec.parameters);

    return querySpec;
}

function retryOrFail(retryPolicy, error, retries, stop, retry) {

    // Stop if error is not retriable.
    if (!error || error.code != 429 || error.retryAfterInMilliseconds === undefined) {
        return stop(error);
    }

    retries = retries || 0;

    // Stop if retry policy is not defined or if there are no retries left.
    if (!retryPolicy || retries >= retryPolicy.max) {
        return stop(error);
    }

    debug('Will try again in %jms (%j of total %j retries).', error.retryAfterInMilliseconds, retries + 1, retryPolicy.max);

    // Retry after given milliseconds.
    setTimeout(function() {
        retry(retries + 1);
    }, error.retryAfterInMilliseconds);
}
